#! /usr/bin/python3 -B

import glob
import os
import re
import shutil
import subprocess
import sys
import tempfile

from dkms_helper import (dkms_modules, get_manual_dkms_output_folder)
from flavour_finder import find_flavours_from_arch


def copy_ko_files_in_tar_folder(module:str, signing_dir_version: str,
                                ko_modules_dir: str):
    module_dir = signing_dir_version + "/" + module
    os.makedirs(module_dir, exist_ok=True)
    for file in os.listdir(ko_modules_dir):
        if file.endswith(".ko"):
            print("DDD --- Copy: " + ko_modules_dir + "/" + file)
            shutil.copy(ko_modules_dir + "/" + file, module_dir + "/" + file)
        # Some installations seems to use DKMS compression, others don't, so
        # check both cases
        elif file.endswith(".ko.zst"):
            print("DDD --- Decomp/copy: " + ko_modules_dir + "/" + file)
            decompressed_file=file.replace(".zst", "")
            with tempfile.TemporaryDirectory() as tmpdir:
                zstd_command = [ "zstd", "-d", ko_modules_dir + "/" + file,
                                "-o", tmpdir + "/" + decompressed_file]
                decompress = subprocess.run(zstd_command,
                              stdout = subprocess.PIPE,
                              stderr = subprocess.PIPE,
                              check = True,
                              text = True)
                shutil.copy(tmpdir + "/" + decompressed_file, module_dir + "/" + decompressed_file)
        else:
            print("DDD --- Unknown file " + file + " Found in ko directory...skipping")


def find_path_build_depends_kos(module, kernel_abi: str, kernel_flavour: str):
    dkms_build_pattern = os.path.join(
            "/var/lib/dkms/",
            module.modulename,
            "*",
            kernel_abi + "-" + kernel_flavour,
            "*",
            "module"
    )

    matches = glob.glob(dkms_build_pattern)
    if not matches:
        raise OSError("Could not find DKMS build artifacts at: " + dkms_build_pattern)
    build_dir = matches[0]
    return build_dir

def prepare_per_flavour(deb_host_arch:str, kernel_version: str,
                        kernel_abi: str, signing_dir: str, flavour: str):
    modules = dkms_modules()
    modules.parse_dkms_version_file()
    modules.filter_per_architecture(deb_host_arch)

    signing_dir_version = signing_dir + "/" + kernel_version
    os.makedirs(signing_dir_version, exist_ok=True)

    for module in modules.items:
        print("DDD - Copying " + module.modulename + " - " + kernel_abi + "-" + flavour)
        ko_src_modules_dir = ""
        ko_dst_module_dir = flavour + "/" + module.getKernelTargetDirectory()

        if not module.needs_off_series:
            # Build-Depends modules path
            print("DDD - Build-depends copy procedure")
            try:
                ko_src_modules_dir = find_path_build_depends_kos(module, kernel_abi, flavour)
            except:
                print("------------------------WARNING------------------------")
                print("------------------------WARNING------------------------")
                print("WARNING: No modules built for the " + flavour + " flavour: skipping copy phase")
                print("------------------------WARNING------------------------")
                print("------------------------WARNING------------------------")
                continue
        else:
            # Manually built modules path
            print("DDD - Off-series copy procedure")
            faketree_src_modules_dir = "/lib/modules/" + kernel_abi + "-" + flavour + "/updates/dkms/"
            ko_src_modules_dir = get_manual_dkms_output_folder() + faketree_src_modules_dir
            ko_src_modules_dir += module.getKernelTargetDirectory()
            if not os.path.exists(ko_src_modules_dir):
                print("------------------------WARNING------------------------")
                print("------------------------WARNING------------------------")
                print("WARNING: No modules built for the " + flavour + " flavour: skipping copy phase")
                print("------------------------WARNING------------------------")
                print("------------------------WARNING------------------------")
                continue

        copy_ko_files_in_tar_folder(ko_dst_module_dir,
                                    signing_dir_version,
                                    ko_src_modules_dir)



def main(deb_host_arch:str, kernel_version: str,
         signing_dir: str, kernel_abi: str):
    flavours = find_flavours_from_arch(deb_host_arch)
    flav_text = ""
    for flavour in flavours:
        flav_text += flavour + " "
    print("[dkms-archive.py] Running with the following parameters:")
    print("   deb_host_arch  = " + deb_host_arch)
    print("   kernel_version = " + kernel_version)
    print("   kernel_abi     = " + kernel_abi)
    print("   signing_dir    = " + signing_dir)
    print("   flavours       = " + flav_text)

    for flavour in flavours:
        prepare_per_flavour(deb_host_arch, kernel_version,
                            kernel_abi, signing_dir, flavour)

# --------------------------------------------------------------------
# deb_host_arh: the archoitecture we are building for (amd64, arm64...)
# kernel_version: the full kernel version (6.8.0-88.89+52)
# signing_dir: the directory whre the ko files will be saved (/SIGNING)
# ko_modules_dir: where the ko files have been built
#               (/lib/modules/$(kernel_version)/updates/dkms/)
(arg_deb_host_arch, arg_kernel_version,
 arg_signing_dir, arg_kernel_abi) = sys.argv[1:]
main(arg_deb_host_arch, arg_kernel_version, arg_signing_dir, arg_kernel_abi)



#! /usr/bin/python3 -B

import re
import sys
import shutil

from dkms_helper import dkms_modules
from flavour_finder import find_flavours
from variant_helper import find_variants

if sys.version_info >= (3, 9):
    List = list
else:
    from typing import List


PACKAGE_BASE_NAME="linux-main-modules"

default_package_template = """
Package: """ + PACKAGE_BASE_NAME + """-@DKMS_NAME@-@KERNEL_ABI@-@KERNEL_FLAVOUR@
Build-Profiles: <!stage1>
Architecture: @ARCHITECTURE@
Section: kernel
Priority: optional
@PROVIDES@
Depends: ${misc:Depends}, ${shlibs:Depends}, kmod, linux-image-@ABI@-@KERNEL_FLAVOUR@ | linux-image-unsigned-@ABI@-@KERNEL_FLAVOUR@
Description: Signed @DKMS_NAME@ module for @KERNEL_ABI@
 This package contains the signed @DKMS_NAME@ modules.
 .
 You likely do not want to install this package directly, Instead, install
 the linux-@DKMS_NAME@-@KERNEL_FLAVOUR@ meta-package, which will ensure that
 upgrades work correctly, and that supporting packages are also installed.
"""

default_variant_package_template = """
Package: linux-modules-@DKMS_NAME@-@KERNEL_FLAVOUR@@KERNEL_VARIANT@
Build-Profiles: <!stage1>
Architecture: @ARCHITECTURE@
Section: kernel
Priority: optional
Depends: ${misc:Depends}, ${shlibs:Depends}, kmod, """ + PACKAGE_BASE_NAME + """-@DKMS_NAME@-@KERNEL_ABI@-@KERNEL_FLAVOUR@
Description: Extra drivers for @DKMS_NAME@ for the @KERNEL_FLAVOUR@ flavour
 Install extra signed @DKMS_NAME@ modules compatible with the @KERNEL_FLAVOUR@ flavour
"""


def create_new_variant_package(dkms_name:str, archs:str, kernel_flavour:str, kernel_abi:str):
    control_variants=""
    variants = find_variants()
    for variant in variants:
        template = default_variant_package_template
        template = re.sub("@DKMS_NAME@", dkms_name, template)
        template = re.sub("@KERNEL_ABI@", kernel_abi, template)
        template = re.sub("@KERNEL_FLAVOUR@", kernel_flavour, template)
        template = re.sub("@ARCHITECTURE@", archs, template)
        template = re.sub("@KERNEL_VARIANT@", variant, template)
        template += "\n"
        control_variants+=template
    return control_variants


def intersect_archs(kernel_archs: List[str], dkms_archs: List[str]):
    res = ""
    for k_arch in kernel_archs:
        if k_arch in dkms_archs:
            if k_arch not in res:
                res += k_arch + " "
    return res.strip()


def create_new_packages(modules, kernel_arch: str,
                        kernel_abi: str, kernel_flavours: List[str]):
    full_template = "\n"
    for kernel_flavour in kernel_flavours:
        for item in modules.items:
            archs = intersect_archs(kernel_flavour.archs, item.arch)
            # If no architectures in common between DKMS and flavour, just skip
            if archs == "":
                continue
            template = default_package_template
            template = re.sub("@DKMS_NAME@", item.modulename, template)
            template = re.sub("@KERNEL_ABI@", kernel_abi, template)
            template = re.sub("@KERNEL_FLAVOUR@", kernel_flavour.flavour, template)
            template = re.sub("@ARCHITECTURE@", archs, template)
            if len(item.rprovides) > 0:
                template = re.sub("@PROVIDES@", "Provides: " + item.getProvidedDkms(), template)
            else:
                template = re.sub("@PROVIDES@\n", "", template)
            template += "\n"
            full_template += template
            ##### This part is only for the variant, to be removed in the future
            full_template += create_new_variant_package(item.modulename,
                                                        archs,
                                                        kernel_flavour.flavour,
                                                        kernel_abi)
            ####################################################################
    return full_template


def create_temporary_stub_file(pkgs_dkms_modules: str):
    control_file = ""
    with open("debian/control.stub", "r") as src:
        control_file = src.read()
    with open("debian/control.stub.tmp", "w") as dst:
        control_file += pkgs_dkms_modules
        dst.write(control_file)


# ------------------------------------------------------------
(arg_deb_host_arch, arg_kernel_abi) = sys.argv[1:]
modules = dkms_modules()

modules.parse_dkms_version_file()

flavours = find_flavours()
pkgs_dkms_modules = create_new_packages(modules, arg_deb_host_arch,
                                        arg_kernel_abi, flavours)
create_temporary_stub_file(pkgs_dkms_modules)



#! /usr/bin/python3 -B

import os
import re
import shutil
import sys
if sys.version_info >= (3, 9):
    List = list
else:
    from typing import List

from dkms_helper import dkms_modules
from flavour_finder import find_flavours

def insert_requirements_in_control_file(modules, kernel_abi_version: str,
                                        kernel_main_version:str, flavours: List[str]):
    res = ""
    for flavour in flavours:
        res += " linux-headers-" + kernel_abi_version + "-" + flavour.flavour
        res += " (>= " + kernel_main_version + ") "
        res += " ["
        archs_size = len(flavour.archs)
        for i, arch in enumerate(flavour.archs):
            res += arch
            if i < archs_size - 1:
                res += " "
        res += "],\n"

    prereqs = modules.return_prerequisites()

    if not prereqs == "":
        res += prereqs
    else:
        res = res.rstrip("\n").rstrip(",")


    final_control = ""
    with open("debian/control", "r") as src:
        control = src.read()
        final_control = re.sub("@DKMS-DEPENDS@", res, control)
    with open("debian/control.tmp", "w") as dst:
        dst.write(final_control)
    shutil.move("debian/control.tmp", "debian/control")


# ------------------------------------------------------
(kernel_abi_version, kernel_main_version) = sys.argv[1:]
modules = dkms_modules()

modules.parse_dkms_version_file()
modules.filter_per_build_depends_build()
# For Debugging purposes
#print("\nModules to be produced:")
#modules.print_modules()
flavours = find_flavours()
insert_requirements_in_control_file(modules, kernel_abi_version, kernel_main_version, flavours)
